###############################################################################
# Solver
# A solver maintains a companion smtlib capable process connected via stdio.
# It can be in 4 main states: None, unknown, sat, unsat
# None      nothing was yet sent to the smtlib process. Al the state is only at the python side
# unknown   is an error state. Some query sent earlier was unsuccessful or timed out.
#           Further interaction with the smtlib process will probably keep returning
#           unknown. An exception is raised.
# sat       the current set of constraints is satisfiable and has at least one solution
# unsat     the current set of constraints is impossible
#
# It starts at None.
# Once you Solver.check() it the status is changed to sat or unsat (or unknown+exception)
# You can create new symbols operate on them. The declarations will be sent to the smtlib process when needed.
# You can add new constraints. A new constraint may change the state from {None, sat} to {sat, unsat, unknown}

import collections
import fcntl
import os
import shlex
import shutil
import time
from abc import abstractmethod
from random import shuffle
from subprocess import PIPE, Popen, check_output
from typing import Any, Sequence, List

from . import operators as Operators
from .constraints import *
from .visitors import *
from ...exceptions import SolverError, SolverUnknown, TooManySolutions, SmtlibError
from ...utils import config

logger = logging.getLogger(__name__)
consts = config.get_group("smt")
consts.add("timeout", default=120, description="Timeout, in seconds, for each Z3 invocation")
consts.add("memory", default=1024 * 8, description="Max memory for Z3 to use (in Megabytes)")
consts.add(
    "maxsolutions",
    default=10000,
    description="Maximum solutions to provide when solving for all values",
)
consts.add("z3_bin", default="z3", description="Z3 solver binary to use")
consts.add("cvc4_bin", default="cvc4", description="CVC4 solver binary to use")
consts.add("yices_bin", default="yices-smt2", description="Yices solver binary to use")
consts.add("boolector_bin", default="boolector", description="Boolector solver binary to use")


consts.add("defaultunsat", default=True, description="Consider solver timeouts as unsat core")
consts.add(
    "optimize", default=True, description="Use smtlib command optimize to find min/max if available"
)

# Regular expressions used by the solver
RE_GET_EXPR_VALUE_ALL = re.compile(
    r"\(([a-zA-Z0-9_]*)[ \n\s]*(#b[0-1]*|#x[0-9a-fA-F]*|[(]?_ bv[0-9]* [0-9]*|true|false)\)"
)
RE_GET_EXPR_VALUE_FMT_BIN = re.compile(r"\(\((?P<expr>(.*))[ \n\s]*#b(?P<value>([0-1]*))\)\)")
RE_GET_EXPR_VALUE_FMT_DEC = re.compile(r"\(\((?P<expr>(.*))\ \(_\ bv(?P<value>(\d*))\ \d*\)\)\)")
RE_GET_EXPR_VALUE_FMT_HEX = re.compile(r"\(\((?P<expr>(.*))\ #x(?P<value>([0-9a-fA-F]*))\)\)")
RE_OBJECTIVES_EXPR_VALUE = re.compile(
    r"\(objectives.*\((?P<expr>.*) (?P<value>\d*)\).*\).*", re.MULTILINE | re.DOTALL
)
RE_MIN_MAX_OBJECTIVE_EXPR_VALUE = re.compile(r"(?P<expr>.*?)\s+\|->\s+(?P<value>.*)", re.DOTALL)

SOLVER_STATS = {"unknown": 0, "timeout": 0}


class SolverType(config.ConfigEnum):
    """Used as configuration constant for choosing solver flavor"""

    z3 = "z3"
    cvc4 = "cvc4"
    yices = "yices"
    auto = "auto"
    portfolio = "portfolio"
    boolector = "boolector"


consts.add(
    "solver",
    default=SolverType.auto,
    description="Choose default smtlib2 solver (z3, yices, cvc4, boolector, portfolio, auto)",
)


def _convert(v):
    r = None
    if v == "true":
        r = True
    elif v == "false":
        r = False
    elif v.startswith("#b"):
        r = int(v[2:], 2)
    elif v.startswith("#x"):
        r = int(v[2:], 16)
    elif v.startswith("_ bv"):
        r = int(v[len("_ bv") : -len(" 256")], 10)
    elif v.startswith("(_ bv"):
        v = v[len("(_ bv") :]
        r = int(v[: v.find(" ")], 10)

    assert r is not None
    return r


class SingletonMixin(object):
    __singleton_instances: Dict[Tuple[int, int], "SingletonMixin"] = {}

    @classmethod
    def instance(cls):
        tid = threading.get_ident()
        pid = os.getpid()
        if not (pid, tid) in cls.__singleton_instances:
            cls.__singleton_instances[(pid, tid)] = cls()
        return cls.__singleton_instances[(pid, tid)]


class SolverException(SmtlibError):
    """
    Solver exception
    """

    pass


class Solver(SingletonMixin):
    def __init__(self):
        pass

    def optimize(self, constraints, X, operation, M=10000):
        """
        Iteratively finds the maximum or minimal value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: the constraints set
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        raise SmtlibError("Abstract method not implemented")

    def check(self, constraints) -> bool:
        """Check if given constraints can be valid"""
        return self.can_be_true(constraints, True)

    def can_be_true(self, constraints, expression=True) -> bool:
        """Check if given expression could be valid"""
        raise SolverException("Abstract method not implemented")

    def must_be_true(self, constraints, expression) -> bool:
        """Check if expression is True and that it can not be False with current constraints"""
        solutions = self.get_all_values(constraints, expression, maxcnt=2, silent=True)
        return solutions == [True]

    def get_all_values(self, constraints, x, maxcnt=10000, silent=False):
        """Returns a list with all the possible values for the symbol x"""
        raise SolverException("Abstract method not implemented")

    def get_value(self, constraints, expression):
        """Ask the solver for one possible result of given expression using given set of constraints."""
        raise SolverException("Abstract method not implemented")

    def max(self, constraints, X: BitVec, M=10000):
        """
        Iteratively finds the maximum value for a symbol within given constraints.
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        assert isinstance(X, BitVec)
        return self.optimize(constraints, X, "maximize", M)

    def min(self, constraints, X: BitVec, M=10000):
        """
        Iteratively finds the minimum value for a symbol within given constraints.

        :param constraints: constraints that the expression must fulfil
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        assert isinstance(X, BitVec)
        return self.optimize(constraints, X, "minimize", M)

    def minmax(self, constraints, x, iters=10000):
        """Returns the min and max possible values for x within given constraints"""
        if issymbolic(x):
            m = self.min(constraints, x, iters)
            M = self.max(constraints, x, iters)
            return m, M
        else:
            return x, x


Version = collections.namedtuple("Version", "major minor patch")


class SmtlibProc:
    def __init__(self, command: str, debug: bool = False):
        """Single smtlib interactive process

        :param command: the shell command to execute
        :param debug: log all messaging
        """
        self._proc: Optional[Popen] = None
        self._command = command
        self._debug = debug
        self._last_buf = ""

    def start(self):
        """Spawns POpen solver process"""
        if self._proc is not None:
            return
        self._proc = Popen(
            shlex.split(self._command),
            stdin=PIPE,
            stdout=PIPE,
            # bufsize=0,  # if we set input to unbuffered, we get syntax errors in large formulas
            universal_newlines=True,
            close_fds=True,
        )

        # stdout should be non-blocking
        fl = fcntl.fcntl(self._proc.stdout, fcntl.F_GETFL)
        fcntl.fcntl(self._proc.stdout, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        self._last_buf = ""

    def stop(self):
        """
        Stops the solver process by:
        - sending a SIGKILL signal,
        - waiting till the process terminates (so we don't leave a zombie process)
        """
        if self._proc is None:
            return
        # if it did not finished already
        if self._proc.returncode is None:
            self._proc.stdin.close()
            self._proc.stdout.close()
            # Kill the process
            self._proc.kill()
            self._proc.wait()

            # No need to wait for termination, zombies avoided.
        self._proc = None

    def send(self, cmd: str) -> None:
        """
        Send a string to the solver.

        :param cmd: a SMTLIBv2 command (ex. (check-sat))
        """
        if self._debug:
            logger.debug(">%s", cmd)
        assert self._proc is not None
        try:
            self._proc.stdout.flush()  # type: ignore
            self._proc.stdin.write(f"{cmd}\n")  # type: ignore
            self._proc.stdin.flush()  # type: ignore
        except (BrokenPipeError, IOError) as e:
            logger.critical(
                f"Solver encountered an error trying to send commands: {e}.\n"
                f"\tOutput: {self._proc.stdout}\n\n"
                f"\tStderr: {self._proc.stderr}"
            )
            raise e

    def recv(self, wait=True) -> Optional[str]:
        """Reads the response from the smtlib solver

        :param wait: a boolean that indicate to wait with a blocking call
        until the results are available. Otherwise, it returns None if the solver
        does not respond.

        """
        tries = 0
        timeout = 0.0

        buf = ""
        if self._last_buf != "":  # we got a partial response last time, let's use it
            buf = buf + self._last_buf

        while True:
            try:
                buf = buf + self._proc.stdout.read()  # type: ignore
                buf = buf.strip()
            except TypeError:
                if not wait:
                    if buf != "":  # we got an error, but something was returned, let's save it
                        self._last_buf = buf
                    return None
                else:
                    tries += 1

            if buf == "":
                continue

            # this verifies if the response from the solver is complete (it has balanced brackets)
            lparen, rparen = map(sum, zip(*((c == "(", c == ")") for c in buf)))
            if lparen == rparen and buf != "":
                break

            if tries > 3:
                time.sleep(timeout)
                timeout += 0.1

        buf = buf.strip()
        self._last_buf = ""

        if "(error" in buf:
            raise SolverException(f"Solver error: {buf}")

        if self._debug:
            logger.debug("<%s", buf)

        return buf

    def _restart(self) -> None:
        """Auxiliary to start or restart the external solver"""
        self.stop()
        self.start()

    def is_started(self):
        return self._proc is not None

    def clear_buffers(self):
        self._proc.stdout.flush()
        self._proc.stdin.flush()


class SMTLIBSolver(Solver):
    ncores: Optional[int] = None
    sname: Optional[str] = None

    @classmethod
    @abstractmethod
    def command(self) -> str:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def inits(self) -> List[str]:
        raise NotImplementedError()

    def __init__(
        self,
        command: str,
        init: Sequence[str] = None,
        support_reset: bool = False,
        support_minmax: bool = False,
        support_pushpop: bool = False,
        multiple_check: bool = True,
        debug: bool = False,
    ):

        """
        Build a smtlib solver instance.
        This is implemented using an external solver (via a subprocess).
        """
        super().__init__()
        self._smtlib: SmtlibProc = SmtlibProc(command, debug)

        # Commands used to initialize smtlib
        if init is None:
            init = tuple()
        self._init = init
        self._support_minmax = support_minmax
        self._support_reset = support_reset
        self._support_pushpop = support_pushpop
        self._multiple_check = multiple_check

        if not self._support_pushpop:
            setattr(self, "_push", None)
            setattr(self, "_pop", None)

        if self._support_minmax and consts.optimize:
            setattr(self, "optimize", self._optimize_fancy)
        else:
            setattr(self, "optimize", self._optimize_generic)

        self._smtlib.start()
        # run solver specific initializations

    def _reset(self, constraints: Optional[str] = None) -> None:
        """Auxiliary method to reset the smtlib external solver to initial defaults"""
        if self._support_reset:
            self._smtlib.start()  # does not do anything if already started
            self._smtlib.send("(reset)")
        else:
            self._smtlib.stop()  # does not do anything if already stopped
            self._smtlib.start()

        self._smtlib.clear_buffers()

        for cfg in self._init:
            self._smtlib.send(cfg)

        if constraints is not None:
            self._smtlib.send(constraints)

    # UTILS: check-sat get-value
    def _is_sat(self) -> bool:
        """
        Check the satisfiability of the current state

        :return: whether current state is satisfiable or not.
        """
        start = time.time()
        self._smtlib.send("(check-sat)")
        status = self._smtlib.recv()
        assert status is not None
        logger.debug("Check took %s seconds (%s)", time.time() - start, status)
        if "ALARM TRIGGERED" in status:
            return False

        if status not in ("sat", "unsat", "unknown"):
            raise SolverError(status)
        if consts.defaultunsat:
            if status == "unknown":
                logger.info("Found an unknown core, probably a solver timeout")
                SOLVER_STATS["timeout"] += 1
                status = "unsat"
                raise SolverUnknown(status)

        if status == "unknown":
            SOLVER_STATS["unknown"] += 1
            raise SolverUnknown(status)
        else:
            assert self.sname is not None
            SOLVER_STATS.setdefault(self.sname, 0)
            SOLVER_STATS[self.sname] += 1

        return status == "sat"

    def _assert(self, expression: Bool):
        """Auxiliary method to send an assert"""
        smtlib = translate_to_smtlib(expression)
        self._smtlib.send(f"(assert {smtlib})")

    def __getvalue_bv(self, expression_str: str) -> int:
        self._smtlib.send(f"(get-value ({expression_str}))")
        t = self._smtlib.recv()
        assert t is not None
        base = 2
        m = RE_GET_EXPR_VALUE_FMT_BIN.match(t)
        if m is None:
            m = RE_GET_EXPR_VALUE_FMT_DEC.match(t)
            base = 10
        if m is None:
            m = RE_GET_EXPR_VALUE_FMT_HEX.match(t)
            base = 16
        if m is None:
            raise SolverError(f"I don't know how to parse the value {str(t)} from {expression_str}")

        expr, value = m.group("expr"), m.group("value")  # type: ignore
        return int(value, base)

    def __getvalue_bool(self, expression_str):
        self._smtlib.send(f"(get-value ({expression_str}))")
        ret = self._smtlib.recv()
        return {"true": True, "false": False, "#b0": False, "#b1": True}[ret[2:-2].split(" ")[1]]

    def __getvalue_all(self, expressions_str: List[str], is_bv: List[bool]) -> Dict[str, int]:
        all_expressions_str = " ".join(expressions_str)
        self._smtlib.send(f"(get-value ({all_expressions_str}))")
        ret_solver: Optional[str] = self._smtlib.recv()
        assert ret_solver is not None
        return_values = re.findall(RE_GET_EXPR_VALUE_ALL, ret_solver)
        return {value[0]: _convert(value[1]) for value in return_values}

    def _getvalue(self, expression) -> Union[int, bool, bytes]:
        """
        Ask the solver for one possible assignment for given expression using current set of constraints.
        The current set of expressions must be sat.

        NOTE: This is an internal method: it uses the current solver state (set of constraints!).
        """
        if not issymbolic(expression):
            return expression

        if isinstance(expression, Array):
            result = bytearray()
            for c in expression:
                expression_str = translate_to_smtlib(c)
                result.append(self.__getvalue_bv(expression_str))
            return bytes(result)
        else:
            if isinstance(expression, BoolVariable):
                return self.__getvalue_bool(expression.name)
            elif isinstance(expression, BitVecVariable):
                return self.__getvalue_bv(expression.name)

        raise NotImplementedError(
            f"_getvalue only implemented for Bool, BitVec and Array. Got {type(expression)}"
        )

    # push pop
    def _push(self):
        """Pushes and save the current constraint store and state."""
        self._smtlib.send("(push 1)")

    def _pop(self):
        """Recall the last pushed constraint store and state."""
        self._smtlib.send("(pop 1)")

    @lru_cache(maxsize=32)
    def can_be_true(self, constraints: ConstraintSet, expression: Union[bool, Bool] = True) -> bool:
        """Check if two potentially symbolic values can be equal"""
        if isinstance(expression, bool):
            if not expression:
                return expression
            else:
                # if True check if constraints are feasible
                self._reset(constraints.to_string())
                return self._is_sat()

        with constraints as temp_cs:
            temp_cs.add(expression)
            self._reset(temp_cs.to_string())
            return self._is_sat()

    # get-all-values min max minmax
    def _optimize_generic(self, constraints: ConstraintSet, x: BitVec, goal: str, max_iter=10000):
        """
        Iteratively finds the maximum or minimum value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: constraints to take into account
        :param x: a symbol or expression
        :param goal: goal to achieve, either 'maximize' or 'minimize'
        :param max_iter: maximum number of iterations allowed
        """
        # TODO: consider adding a mode to return best known value on timeout
        assert goal in ("maximize", "minimize")
        operation = {"maximize": Operators.UGE, "minimize": Operators.ULE}[goal]

        last_value: Optional[Union[int, bool, bytes]] = None

        start = time.time()
        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)  # _getvalue needs a Variable
            temp_cs.add(X == x)
            self._reset(temp_cs.to_string())

            # Find one value and use it as currently known min/Max
            if not self._is_sat():
                raise SolverException("UNSAT")
            last_value = self._getvalue(X)
            self._assert(operation(X, last_value))

            # This uses a binary search to find a suitable range for aux
            # Use known solution as min or max depending on the goal
            if goal == "maximize":
                m, M = last_value, (1 << X.size) - 1
            else:
                m, M = 0, last_value

            # Iteratively divide the range
            L = None
            while L not in (M, m):
                L = (m + M) // 2
                self._assert(operation(X, L))
                sat = self._is_sat()

                # depending on the goal move one of the extremes
                if goal == "maximize" and sat or goal == "minimize" and not sat:
                    m = L
                else:
                    M = L

                if time.time() - start > consts.timeout:
                    SOLVER_STATS["timeout"] += 1
                    raise SolverError("Timeout")

        # reset to before the dichotomic search
        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)  # _getvalue needs a Variable
            temp_cs.add(X == x)
            self._reset(temp_cs.to_string())

            # At this point we know aux is inside [m,M]
            # Lets constrain it to that range
            self._assert(Operators.UGE(X, m))
            self._assert(Operators.ULE(X, M))

            # And now check all remaining possible extremes
            last_value = None
            i = 0
            while self._is_sat():
                last_value = self._getvalue(X)
                self._assert(operation(X, last_value))
                self._assert(X != last_value)
                i = i + 1
                if i > max_iter:
                    SOLVER_STATS["unknown"] += 1
                    raise SolverError("Optimizing error, maximum number of iterations was reached")
                if time.time() - start > consts.timeout:
                    SOLVER_STATS["timeout"] += 1
                    raise SolverError("Timeout")
            if last_value is not None:
                return last_value

            SOLVER_STATS["unknown"] += 1
            raise SolverError("Optimizing error, unsat or unknown core")

    @lru_cache(maxsize=32)
    def get_all_values(
        self,
        constraints: ConstraintSet,
        expression,
        maxcnt: Optional[int] = None,
        silent: bool = False,
    ):
        """Returns a list with all the possible values for the symbol x"""
        if not isinstance(expression, Expression):
            return [expression]
        assert isinstance(expression, Expression)
        expression = simplify(expression)
        if maxcnt is None:
            maxcnt = consts.maxsolutions
            if isinstance(expression, Bool) and consts.maxsolutions > 1:
                # We know there is max 2 solutions when Bool
                maxcnt = 2
                silent = True

        with constraints as temp_cs:
            if isinstance(expression, Bool):
                var = temp_cs.new_bool()
            elif isinstance(expression, BitVec):
                var = temp_cs.new_bitvec(expression.size)
            elif isinstance(expression, Array):
                var = temp_cs.new_array(
                    index_max=expression.index_max,
                    value_bits=expression.value_bits,
                    taint=expression.taint,
                ).array
            else:
                raise NotImplementedError(
                    f"get_all_values only implemented for {type(expression)} expression type."
                )

            temp_cs.add(var == expression)
            self._reset(temp_cs.to_string())
            result = []
            start = time.time()
            while self._is_sat():
                value = self._getvalue(var)
                result.append(value)

                if len(result) >= maxcnt:
                    if silent:
                        # do not throw an exception if set to silent
                        # Default is not silent, assume user knows
                        # what they are doing and will check the size
                        # of returned vals list (previous smtlib behavior)
                        break
                    else:
                        raise TooManySolutions(result)
                if time.time() - start > consts.timeout:
                    SOLVER_STATS["timeout"] += 1
                    if silent:
                        logger.info("Timeout searching for all solutions")
                        return list(result)
                    raise SolverError("Timeout")
                # Sometimes adding a new contraint after a check-sat eats all the mem
                if self._multiple_check:
                    self._smtlib.send(f"(assert {translate_to_smtlib(var != value)})")
                else:
                    temp_cs.add(var != value)
                    self._reset(temp_cs.to_string())
            return list(result)

    def _optimize_fancy(self, constraints: ConstraintSet, x: BitVec, goal: str, max_iter=10000):
        """
        Iteratively finds the maximum or minimum value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: constraints to take into account
        :param x: a symbol or expression
        :param goal: goal to achieve, either 'maximize' or 'minimize'
        :param max_iter: maximum number of iterations allowed
        """
        # TODO: consider adding a mode to return best known value on timeout
        assert goal in ("maximize", "minimize")
        operation = {"maximize": Operators.UGE, "minimize": Operators.ULE}[goal]

        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)
            temp_cs.add(X == x)
            aux = temp_cs.new_bitvec(X.size, name="optimized_")
            self._reset(temp_cs.to_string())

            self._assert(operation(X, aux))
            self._smtlib.send("(%s %s)" % (goal, aux.name))
            self._smtlib.send("(check-sat)")
            _status = self._smtlib.recv()

            assert self.sname is not None
            SOLVER_STATS.setdefault(self.sname, 0)
            SOLVER_STATS[self.sname] += 1

            if _status == "sat":
                return self._getvalue(aux)
            raise SolverError("Optimize failed")

    def get_value(self, constraints: ConstraintSet, *expressions):
        values = self.get_value_in_batch(constraints, expressions)
        if len(expressions) == 1:
            return values[0]
        else:
            return values

    def get_value_in_batch(self, constraints: ConstraintSet, expressions):
        """
        Ask the solver for one possible result of given expressions using
        given set of constraints.
        """
        values: List[Any] = [None] * len(expressions)
        start = time.time()
        with constraints.related_to(*expressions) as temp_cs:
            vars: List[Any] = []
            for idx, expression in enumerate(expressions):
                if not issymbolic(expression):
                    values[idx] = expression
                    vars.append(None)
                    continue
                assert isinstance(expression, (Bool, BitVec, Array))
                if isinstance(expression, Bool):
                    var = temp_cs.new_bool()
                    vars.append(var)
                    temp_cs.add(var == expression)
                elif isinstance(expression, BitVec):
                    var = temp_cs.new_bitvec(expression.size)
                    vars.append(var)
                    temp_cs.add(var == expression)
                elif isinstance(expression, Array):
                    var = []
                    for i in range(expression.index_max):
                        subvar = temp_cs.new_bitvec(expression.value_bits)
                        var.append(subvar)
                        temp_cs.add(subvar == simplify(expression[i]))
                    vars.append(var)

            self._reset(temp_cs.to_string())
            if not self._is_sat():
                raise SolverError(
                    "Solver could not find a value for expression under current constraint set"
                )

            values_to_ask: List[str] = []
            is_bv: List[bool] = []
            for idx, expression in enumerate(expressions):
                if not issymbolic(expression):
                    continue
                var = vars[idx]
                if isinstance(expression, Bool):
                    values_to_ask.append(var.name)
                    is_bv.append(False)
                if isinstance(expression, BitVec):
                    values_to_ask.append(var.name)
                    is_bv.append(True)
                if isinstance(expression, Array):
                    # result = []
                    for i in range(expression.index_max):
                        values_to_ask.append(var[i].name)
                        is_bv.append(True)

            if values_to_ask == []:
                return values

            values_returned = self.__getvalue_all(values_to_ask, is_bv)
            for idx, expression in enumerate(expressions):
                if not issymbolic(expression):
                    continue
                var = vars[idx]
                if isinstance(expression, Bool):
                    values[idx] = values_returned[var.name]
                if isinstance(expression, BitVec):
                    if var.name not in values_returned:
                        logger.error(
                            "var.name", var.name, "not in values_returned", values_returned
                        )

                    values[idx] = values_returned[var.name]
                if isinstance(expression, Array):
                    result = []
                    for i in range(expression.index_max):
                        result.append(values_returned[var[i].name])
                    values[idx] = bytes(result)

            if time.time() - start > consts.timeout:
                SOLVER_STATS["timeout"] += 1
                raise SolverError("Timeout")

        return values


class Z3Solver(SMTLIBSolver):
    sname = "z3"

    @classmethod
    def command(self) -> str:
        return f"{consts.z3_bin} -t:{consts.timeout * 1000} -memory:{consts.memory} -smt2 -in"

    @classmethod
    def inits(self) -> List[str]:
        return [
            "(set-logic QF_AUFBV)",
            "(set-option :global-decls false)",
            "(set-option :tactic.solve_eqs.context_solve false)",
        ]

    def __init__(self):
        """
        Build a Z3 solver instance.
        This is implemented using an external z3 solver (via a subprocess).
        See https://github.com/Z3Prover/z3
        """
        command = self.command()
        self.ncores = 1

        init, support_minmax, support_reset, multiple_check = self.__autoconfig()
        super().__init__(
            command=command,
            init=init,
            support_minmax=support_minmax,
            support_reset=support_reset,
            multiple_check=multiple_check,
            support_pushpop=True,
            debug=False,
        )

    def __autoconfig(self):
        init = self.inits()

        # To cache what get-info returned; can be directly set when writing tests
        self.version = self._solver_version()
        support_reset = True

        support_minmax = self.version >= Version(4, 4, 1)

        # Certain version of Z3 fails to handle multiple check-sat
        # https://gist.github.com/feliam/0f125c00cb99ef05a6939a08c4578902
        multiple_check = self.version < Version(4, 8, 7)
        return init, support_minmax, support_reset, multiple_check

    def _solver_version(self) -> Version:
        """
        If we fail to parse the version, we assume z3's output has changed, meaning it's a newer
        version than what's used now, and therefore ok.

        Anticipated version_cmd_output format: 'Z3 version 4.4.2'
                                               'Z3 version 4.4.5 - 64 bit - build hashcode $Z3GITHASH'
        """
        try:
            received_version = check_output([f"{consts.z3_bin}", "--version"])
            Z3VERSION = re.compile(
                r".*(?P<major>([0-9]+))\.(?P<minor>([0-9]+))\.(?P<patch>([0-9]+)).*"
            )
            m = Z3VERSION.match(received_version.decode("utf-8"))
            major, minor, patch = map(
                int, (m.group("major"), m.group("minor"), m.group("patch"))  # type: ignore
            )
            parsed_version = Version(major, minor, patch)
        except (ValueError, TypeError) as e:
            logger.warning(
                f"Could not parse Z3 version: '{str(received_version)}'. Assuming compatibility."
            )
            parsed_version = Version(float("inf"), float("inf"), float("inf"))
        return parsed_version


class YicesSolver(SMTLIBSolver):
    sname = "yices"

    @classmethod
    def command(self) -> str:
        return f"{consts.yices_bin} --timeout={consts.timeout}  --incremental"

    @classmethod
    def inits(self) -> List[str]:
        return ["(set-logic QF_AUFBV)"]

    def __init__(self):
        init = self.inits()
        command = self.command()
        self.ncores = 1
        super().__init__(
            command=command,
            init=init,
            debug=False,
            support_minmax=False,
            support_reset=False,
        )


class CVC4Solver(SMTLIBSolver):
    sname = "cvc4"

    @classmethod
    def command(self) -> str:
        return f"{consts.cvc4_bin} --tlimit={consts.timeout * 1000} --lang=smt2 --incremental"

    @classmethod
    def inits(self) -> List[str]:
        return ["(set-logic QF_AUFBV)", "(set-option :produce-models true)"]

    def __init__(self):
        init = self.inits()
        command = self.command()
        self.ncores = 1
        super().__init__(command=command, init=init)


class BoolectorSolver(SMTLIBSolver):
    sname = "boolector"

    @classmethod
    def command(self) -> str:
        return f"{consts.boolector_bin} --time={consts.timeout} -i"

    @classmethod
    def inits(self) -> List[str]:
        return ["(set-logic QF_AUFBV)", "(set-option :produce-models true)"]

    def __init__(self, args: List[str] = []):
        init = self.inits()
        command = self.command()
        self.ncores = 1
        super().__init__(command=command, init=init)


class SmtlibPortfolio:
    def __init__(self, solvers: List[str], debug: bool = False):
        """Single smtlib interactive process

        :param command: the shell command to execute
        :param debug: log all messaging
        """
        self._procs: Dict[str, SmtlibProc] = {}
        self._solvers: List[str] = solvers
        self._debug = debug

    def start(self):
        if len(self._procs) == 0:
            for solver in self._solvers:
                self._procs[solver] = SmtlibProc(solver_selector[solver].command(), self._debug)

        for _, proc in self._procs.items():
            proc.start()

    def stop(self):
        """
        Stops the solver process by:
        - sending a SIGKILL signal,
        - waiting till the process terminates (so we don't leave a zombie process)
        """
        for solver, proc in self._procs.items():
            proc.stop()

    def send(self, cmd: str) -> None:
        """
        Send a string to the solver.

        :param cmd: a SMTLIBv2 command (ex. (check-sat))
        """
        assert len(self._procs) > 0
        inds = list(range(len(self._procs)))
        shuffle(inds)

        for i in inds:
            solver = self._solvers[i]
            proc = self._procs[solver]
            if not proc.is_started():
                continue

            proc.send(cmd)

    def recv(self) -> str:
        """Reads the response from the smtlib solver"""
        tries = 0
        timeout = 0.0
        inds = list(range(len(self._procs)))
        # print(self._solvers)
        while True:
            shuffle(inds)
            for i in inds:

                solver = self._solvers[i]
                proc = self._procs[solver]

                if not proc.is_started():
                    continue

                buf = proc.recv(wait=False)
                if buf is not None:

                    for osolver in self._solvers:  # iterate on all the solvers
                        if osolver != solver:  # check for the other ones
                            self._procs[osolver].stop()  # stop them

                    return buf
                else:
                    tries += 1

            if tries > 10 * len(self._procs):
                time.sleep(timeout)
                timeout += 0.1

    def _restart(self) -> None:
        """Auxiliary to start or restart the external solver"""
        self.stop()
        self.start()

    def is_started(self):
        return len(self._procs) > 0

    def init(self):
        assert len(self._solvers) == len(self._procs)
        for solver, proc in self._procs.items():
            for cfg in solver_selector[solver].inits():
                proc.send(cfg)


class PortfolioSolver(SMTLIBSolver):
    sname = "portfolio"

    def __init__(self):
        solvers = []
        if shutil.which(consts.yices_bin):
            solvers.append(consts.solver.yices.name)
        # not sure we want z3 here, since it tends to be slower
        # if shutil.which(consts.z3_bin):
        #    solvers.append(consts.solver.z3.name)
        if shutil.which(consts.cvc4_bin):
            solvers.append(consts.solver.cvc4.name)
        if shutil.which(consts.boolector_bin):
            solvers.append(consts.solver.boolector.name)
        if solvers == []:
            raise SolverException(
                f"No Solver not found. Install one ({consts.yices_bin}, {consts.z3_bin}, {consts.cvc4_bin}, {consts.boolector_bin})."
            )

        logger.info("Creating portfolio with solvers: " + ",".join(solvers))
        assert len(solvers) > 0
        support_reset: bool = False
        support_minmax: bool = False
        support_pushpop: bool = False
        multiple_check: bool = True
        debug: bool = False

        self._smtlib: SmtlibPortfolio = SmtlibPortfolio(solvers, debug)
        self._support_minmax = support_minmax
        self._support_reset = support_reset
        self._support_pushpop = support_pushpop
        self._multiple_check = multiple_check

        if not self._support_pushpop:
            setattr(self, "_push", None)
            setattr(self, "_pop", None)

        if self._support_minmax and consts.optimize:
            setattr(self, "optimize", self._optimize_fancy)
        else:
            setattr(self, "optimize", self._optimize_generic)
        self.ncores = len(solvers)

    def _reset(self, constraints: Optional[str] = None) -> None:
        """Auxiliary method to reset the smtlib external solver to initial defaults"""
        if self._support_reset:
            self._smtlib.start()  # does not do anything if already started
            self._smtlib.send("(reset)")
        else:
            self._smtlib.stop()  # does not do anything if already stopped
            self._smtlib.start()

        self._smtlib.init()

        if constraints is not None:
            self._smtlib.send(constraints)


solver_selector = {
    "cvc4": CVC4Solver,
    "boolector": BoolectorSolver,
    "yices": YicesSolver,
    "z3": Z3Solver,
    "portfolio": PortfolioSolver,
}


class SelectedSolver:
    choice = None

    @classmethod
    def instance(cls):
        if consts.solver == consts.solver.auto:
            if cls.choice is None:
                if shutil.which(consts.yices_bin):
                    cls.choice = consts.solver.yices
                elif shutil.which(consts.z3_bin):
                    cls.choice = consts.solver.z3
                elif shutil.which(consts.cvc4_bin):
                    cls.choice = consts.solver.cvc4
                elif shutil.which(consts.boolector_bin):
                    cls.choice = consts.solver.boolector
                else:
                    raise SolverException(
                        f"No Solver not found. Install one ({consts.yices_bin}, {consts.z3_bin}, {consts.cvc4_bin}, {consts.boolector_bin})."
                    )

        else:
            cls.choice = consts.solver

        SelectedSolver = solver_selector[cls.choice.name]
        return SelectedSolver.instance()



import typing
from dataclasses import dataclass
from ..core.smtlib import issymbolic, BitVec
from ctypes import *
import wasm
import struct
from ..core.state import Concretize

# Do I like this? No. Is it necessary to get mypy to pass without destroying the WASM type system? Yes.
# Look how much cleaner the version below is instead of having a separate class for every type...
"""
U32: type = type("U32", (int,), {})
U64: type = type("U64", (int,), {})
"""


class U32(int):
    pass


class U64(int):
    pass


def debug(imm):
    """
    Attempts to pull meaningful data out of an immediate, which has a dynamic GeneratedStructure type

    :param imm: the instruction immediate
    :return: a printable representation of the immediate, or the immediate itself
    """
    if hasattr(imm, "value"):
        return imm.value
    if hasattr(imm, "function_index"):
        return f"Func Idx {imm.function_index}"
    if hasattr(imm, "offset"):
        return f"Offset {imm.offset}"
    if hasattr(imm, "local_index"):
        return f"Local {imm.local_index}"
    if hasattr(imm, "global_index"):
        return f"Global {imm.global_index}"
    return getattr(imm, "value", imm)


def _reinterpret(ty1: type, ty2: type, val):
    """
    Attempts to convert a value from one ctypes type to another

    :param ty1: The type of the value
    :param ty2: The desired type of the value
    :param val: The value itself
    :return: The converted value
    """
    ptr = pointer(ty1(val))
    # mypy worries that `contents` won't always exist for the pointer type
    return cast(ptr, POINTER(ty2)).contents.value  # type: ignore


class I32(int):
    """
    Subclass of int that's restricted to 32-bit values
    """

    def __new__(cls, val):
        val = struct.unpack("i", c_int32(int(val)))[0]
        return super(I32, cls).__new__(cls, val)

    @classmethod
    def cast(cls, other):
        """
        :param other: Value to convert to I32
        :return: If other is symbolic, other. Otherwise, I32(other)
        """
        if issymbolic(other):
            return other
        return cls(other)

    @staticmethod
    def to_unsigned(val):
        """
        Reinterprets the argument from a signed integer to an unsigned 32-bit integer

        :param val: Signed integer to reinterpret
        :return: The unsigned equivalent
        """
        return _reinterpret(c_int32, c_uint32, val)


class I64(int):
    """
    Subclass of int that's restricted to 64-bit values
    """

    def __new__(cls, val):
        val = struct.unpack("q", c_int64(int(val)))[0]
        return super(I64, cls).__new__(cls, val)

    @classmethod
    def cast(cls, other):
        """
        :param other: Value to convert to I64
        :return: If other is symbolic, other. Otherwise, I64(other)
        """
        if issymbolic(other):
            return other
        return cls(other)

    @staticmethod
    def to_unsigned(val):
        """
        Reinterprets the argument from a signed integer to an unsigned 64-bit integer

        :param val: Signed integer to reinterpret
        :return: The unsigned equivalent
        """
        return _reinterpret(c_int64, c_uint64, val)


class F32(float):
    """
    Subclass of float that's restricted to 32-bit values
    """

    def __new__(cls, val):
        if isinstance(val, int):
            val = _reinterpret(c_int32, c_float, val & 0xFFFFFFFF)
        val = struct.unpack("f", c_float(val))[0]
        self = super(F32, cls).__new__(cls, val)
        self.integer = val
        return self

    @classmethod
    def cast(cls, other):
        """
        :param other: Value to convert to F32
        :return: If other is symbolic, other. Otherwise, F32(other)
        """
        if issymbolic(other):
            return other
        return cls(other)


class F64(float):
    """
    Subclass of float that's restricted to 64-bit values
    """

    def __new__(cls, val):
        if isinstance(val, int):
            val = _reinterpret(c_int64, c_double, val)
        val = struct.unpack("d", c_double(val))[0]
        self = super(F64, cls).__new__(cls, val)
        self.integer = val
        return self

    @classmethod
    def cast(cls, other):
        """
        :param other: Value to convert to F64
        :return: If other is symbolic, other. Otherwise, F64(other)
        """
        if issymbolic(other):
            return other
        return cls(other)


ValType = type  #: https://www.w3.org/TR/wasm-core-1/#syntax-valtype
Value_t = (I32, I64, F32, F64, BitVec)
# Value = typing.TypeVar('Value', I32, I64, F32, F64, BitVec)  #: https://www.w3.org/TR/wasm-core-1/#syntax-val
Value = typing.Union[I32, I64, F32, F64, BitVec]  #: https://www.w3.org/TR/wasm-core-1/#syntax-val


class Name(str):
    pass


@dataclass
class FunctionType:
    """
    https://www.w3.org/TR/wasm-core-1/#syntax-functype
    """

    param_types: typing.List[ValType]  #: Sequential types of each of the parameters
    result_types: typing.List[ValType]  #: Sequential types of each of the return values


@dataclass
class LimitType:
    """
    https://www.w3.org/TR/wasm-core-1/#syntax-limits
    """

    min: U32
    max: typing.Optional[U32]


@dataclass
class TableType:
    """https://www.w3.org/TR/wasm-core-1/#syntax-tabletype"""

    limits: LimitType  #: Minimum and maximum size of the table
    elemtype: type  #: the type ot the element. Currently, the only element type is `funcref`


@dataclass
class GlobalType:
    """https://www.w3.org/TR/wasm-core-1/#syntax-globaltype"""

    mut: bool  #: Whether or not this global is mutable
    value: ValType  #: The value of the global


# https://www.w3.org/TR/wasm-core-1/#indices%E2%91%A0
class TypeIdx(U32):
    pass


class FuncIdx(U32):
    pass


class TableIdx(U32):
    pass


class MemIdx(U32):
    pass


class GlobalIdx(U32):
    pass


class LocalIdx(U32):
    pass


class LabelIdx(U32):
    pass


@dataclass
class BlockImm:
    sig: int


@dataclass
class BranchImm:
    relative_depth: U32


@dataclass
class BranchTableImm:
    target_count: U32
    target_table: typing.List[U32]
    default_target: U32


@dataclass
class CallImm:
    function_index: U32


@dataclass
class CallIndirectImm:
    type_index: U32
    reserved: U32


@dataclass
class LocalVarXsImm:
    local_index: U32


@dataclass
class GlobalVarXsImm:
    global_index: U32


@dataclass
class MemoryImm:
    flags: U32
    offset: U32


@dataclass
class CurGrowMemImm:
    reserved: bool


@dataclass
class I32ConstImm:
    value: I32


@dataclass
class I64ConstImm:
    value: I64


@dataclass
class F32ConstImm:
    value: F32


@dataclass
class F64ConstImm:
    value: F64


ImmType = typing.Union[
    BlockImm,
    BranchImm,
    BranchTableImm,
    CallImm,
    CallIndirectImm,
    LocalVarXsImm,
    GlobalVarXsImm,
    MemoryImm,
    CurGrowMemImm,
    I32ConstImm,
    F32ConstImm,
    F64ConstImm,
]  #: Types of all immediates


class Instruction:
    """Internal instruction class that's pickle-friendly and works with the type system"""

    __slots__ = ["opcode", "mnemonic", "imm"]
    opcode: int  #: Opcode, used for dispatching instructions
    mnemonic: str  #: Used for debugging
    imm: ImmType  #: A class with the immediate data for this instruction

    def __init__(self, inst: wasm.decode.Instruction, imm=None):
        self.opcode = inst.op.id
        self.mnemonic = inst.op.mnemonic
        self.imm = imm

    def __repr__(self):
        return f"<Instruction: {self.mnemonic} ({debug(self.imm)})>"


MemoryType = LimitType  #: https://www.w3.org/TR/wasm-core-1/#syntax-memtype
ExternType = typing.Union[
    FunctionType, TableType, MemoryType, GlobalType
]  #: https://www.w3.org/TR/wasm-core-1/#external-types%E2%91%A0
WASMExpression = typing.List[Instruction]


def convert_instructions(inst_seq) -> WASMExpression:
    """
    Converts instructions output from the parser into full-fledged Python objects that will work with Manticore.
    This is necessary because the pywasm module uses lots of reflection to generate structures on the fly, which
    doesn't play nicely with Pickle or the type system. That's why we need the `debug` method above to print out
    immediates, and also why we've created a separate class for every different type of immediate.

    :param inst_seq: Sequence of raw instructions to process
    :return: The properly-typed instruction sequence in a format Manticore can use
    """
    out = []
    if not isinstance(inst_seq, list):
        inst_seq = list(wasm.decode_bytecode(inst_seq))
    i: wasm.decode.Instruction
    for i in inst_seq:
        if 0x02 <= i.op.id <= 0x04:
            out.append(Instruction(i, BlockImm(i.imm.sig)))
        elif i.op.id in (0x0C, 0x0D):
            out.append(Instruction(i, BranchImm(i.imm.relative_depth)))
        elif i.op.id == 0x0E:
            out.append(
                Instruction(
                    i, BranchTableImm(i.imm.target_count, i.imm.target_table, i.imm.default_target)
                )
            )
        elif i.op.id == 0x10:
            out.append(Instruction(i, CallImm(i.imm.function_index)))
        elif i.op.id == 0x11:
            out.append(Instruction(i, CallIndirectImm(i.imm.type_index, i.imm.reserved)))
        elif 0x20 <= i.op.id <= 0x22:
            out.append(Instruction(i, LocalVarXsImm(i.imm.local_index)))
        elif i.op.id in (0x23, 0x24):
            out.append(Instruction(i, GlobalVarXsImm(i.imm.global_index)))
        elif 0x28 <= i.op.id <= 0x3E:
            out.append(Instruction(i, MemoryImm(i.imm.flags, i.imm.offset)))
        elif i.op.id in (0x3F, 0x40):
            out.append(Instruction(i, CurGrowMemImm(i.imm.reserved)))
        elif i.op.id == 0x41:
            out.append(Instruction(i, I32ConstImm(i.imm.value)))
        elif i.op.id == 0x42:
            out.append(Instruction(i, I64ConstImm(i.imm.value)))
        elif i.op.id == 0x43:
            out.append(Instruction(i, F32ConstImm(i.imm.value)))
        elif i.op.id == 0x44:
            out.append(Instruction(i, F64ConstImm(i.imm.value)))
        else:
            out.append(Instruction(i))

    return out


class Trap(Exception):
    """
    Subclass of Exception, used for WASM errors
    """

    pass


class UnreachableInstructionTrap(Trap):
    def __init__(self):
        super().__init__("Tried to execute an unreachable instruction")


class ZeroDivisionTrap(Trap):
    def __init__(self):
        super().__init__("Zero division")


class OverflowDivisionTrap(Trap):
    def __init__(self):
        super().__init__("Overflow in signed division")


class NonExistentFunctionCallTrap(Trap):
    def __init__(self):
        super().__init__("Indirect call to non-existent function")


class OutOfBoundsMemoryTrap(Trap):
    def __init__(self, addr):
        super().__init__("Out of bounds memory access at " + hex(addr))


class InvalidConversionTrap(Trap):
    def __init__(self, ty, val):
        super().__init__("Can't convert " + str(val) + " to " + str(ty))


class TypeMismatchTrap(Trap):
    def __init__(self, ty1, ty2):
        super().__init__(f"Type signature mismatch: {ty1} != {ty2}")


class ConcretizeStack(Concretize):
    """Tells Manticore to concretize the value `depth` values from the end of the stack."""

    def __init__(self, depth: int, ty: type, message: str, expression, policy=None, **kwargs):
        """
        :param depth: Index in the stack (should typically be negative)
        :param ty: The type to cast the
        :param message: Debug message describing the reason for concretization
        :param expression: The expression to concretize, either a Value or a BitVec
        """

        def setstate(state, value):
            state.platform.stack.data[depth] = ty(value)

        super().__init__(message, expression, setstate, policy, **kwargs)


class MissingExportException(Trap):
    def __init__(self, name):
        self.name = name
        super().__init__(f"Couldn't find an export called `{name}`")



# Minimal INTEL assembler expression calculator
import ply.yacc as yacc
import copy
from ..smtlib import Operators

# Lexer
# ------------------------------------------------------------
# calclex.py
#
# tokenizer for a simple expression evaluator for
# numbers and +,-,*,/
# ------------------------------------------------------------
import ply.lex as lex
import re


class ParserException(Exception):
    """
    Parser exception
    """

    pass


# List of token names.   This is always required
tokens = (
    "NUMBER",
    "PLUS",
    "MINUS",
    "TIMES",
    "DIVIDE",
    "AND",
    "OR",
    "NEG",
    "LPAREN",
    "RPAREN",
    "LBRAKET",
    "RBRAKET",
    "REGISTER",
    "SEGMENT",
    "COLOM",
    "PTR",
    "TYPE",
    "RSHIFT",
    "LSHIFT",
    "LOR",
    "LAND",
    "LNOT",
    "EQ",
    "LT",
    "LE",
    "GT",
    "GE",
)

# Regular expression rules for simple tokens
t_PLUS = r"\+"
t_MINUS = r"-"
t_TIMES = r"\*"
t_DIVIDE = r"/"
t_LPAREN = r"\("
t_RPAREN = r"\)"
t_LBRAKET = r"\["
t_RBRAKET = r"\]"
t_COLOM = r":"

t_AND = r"&"
t_OR = r"\|"
t_NEG = r"~"
t_LSHIFT = r"<<"
t_RSHIFT = r">>"

t_LAND = r"&&"
t_LOR = r"\|\|"
t_LNOT = r"!"

t_EQ = r"=="
t_LT = r"<"
t_LE = r"<="
t_GT = r">"
t_GE = r">="


re_NUMBER = re.compile(r"^(0x[a-fA-F0-9]+|[a-fA-F0-9]+)$")
re_REGISTER = re.compile(
    r"^(EAX|EBX|ECX|EDX|ESI|EDI|ESP|EBP|RAX|RBX|RCX|RDX|RSI|RDI|RSP|RBP|ZF|CF|SF|EFLAGS)$"
)
re_SEGMENT = re.compile(r"^(SS|DS|ES|SS|CS)$")
re_TYPE = re.compile(r"^(QWORD|DWORD|WORD|BYTE)$")
re_PTR = re.compile(r"^PTR$")

# A regular expression rule with some action code


def t_TOKEN(t):
    "[a-zA-Z0-9]+"
    # print t.value,t.lexer.lexdata[t.lexer.lexpos-len(t.value):],re_TYPE.match(t.lexer.lexdata,t.lexer.lexpos-len(t.value))
    if re_TYPE.match(t.value):
        t.type = "TYPE"
    elif re_PTR.match(t.value):
        t.type = "PTR"
    elif re_NUMBER.match(t.value):
        if t.value.startswith("0x"):
            t.value = t.value[2:]
        t.value = int(t.value, 16)
        t.type = "NUMBER"
    elif re_REGISTER.match(t.value):
        t.type = "REGISTER"
    elif re_SEGMENT.match(t.value):
        t.type = "SEGMENT"
    else:
        raise ParserException(f"Unknown:<{t.value}>")
    return t


# Define a rule so we can track line numbers


def t_newline(t):
    r"\n+"
    t.lexer.lineno += len(t.value)


# A string containing ignored characters (spaces and tabs)
t_ignore = " \t"

# Error handling rule


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)


# Build the lexer
lexer = lex.lex()


# parser


precedence = (
    ("left", "PLUS", "MINUS"),
    ("left", "DIVIDE"),
    ("left", "TIMES"),
    ("left", "AND", "OR"),
    ("right", "NEG"),
)


def default_read_memory(address, size):
    return f"READM({address:08x},{size})"


def default_read_register(reg):
    return f"REG({reg})"


def default_get_descriptor(selector):
    return (0, 0xFFFFF000, "rwx")


default_sizes_32 = {"QWORD": 8, "DWORD": 4, "WORD": 2, "BYTE": 1}

default_sizes_64 = {"QWORD": 8, "DWORD": 4, "WORD": 2, "BYTE": 1}
functions = {
    "read_memory": default_read_memory,
    "read_register": default_read_register,
    "get_descriptor": default_get_descriptor,
}
sizes = copy.copy(default_sizes_32)


def p_expression_div(p):
    "expression : expression DIVIDE expression"
    p[0] = p[1] // p[3]


def p_expression_mul(p):
    "expression : expression TIMES expression"
    p[0] = p[1] * p[3]


def p_expression_plus(p):
    "expression : expression PLUS expression"
    p[0] = p[1] + p[3]


def p_expression_minus(p):
    "expression : expression MINUS expression"
    p[0] = p[1] - p[3]


def p_expression_and(p):
    "expression : expression AND expression"
    p[0] = p[1] & p[3]


def p_expression_or(p):
    "expression : expression OR expression"
    p[0] = p[1] | p[3]


def p_expression_neg(p):
    "expression : NEG expression"
    p[0] = ~p[1]


def p_expression_lshift(p):
    "expression : expression LSHIFT expression"
    p[0] = p[1] << p[3]


def p_expression_rshift(p):
    "expression : expression RSHIFT expression"
    p[0] = p[1] >> p[3]


def p_expression_deref(p):
    "expression : TYPE PTR LBRAKET expression RBRAKET"
    size = sizes[p[1]]
    address = p[4]
    char_list = functions["read_memory"](address, size)
    value = Operators.CONCAT(8 * len(char_list), *reversed(map(Operators.ORD, char_list)))
    p[0] = value


def p_expression_derefseg(p):
    "expression : TYPE PTR SEGMENT COLOM LBRAKET expression RBRAKET"
    size = sizes[p[1]]
    address = p[6]
    seg = functions["read_register"](p[3])
    base, limit, _ = functions["get_descriptor"](seg)
    address = base + address
    char_list = functions["read_memory"](address, size)
    value = Operators.CONCAT(8 * len(char_list), *reversed(map(Operators.ORD, char_list)))
    p[0] = value


def p_expression_term(p):
    "expression : term"
    p[0] = p[1]


def p_factor_expr(p):
    "expression : LPAREN expression RPAREN"
    p[0] = p[2]


def p_term_num(p):
    "term : NUMBER"
    p[0] = p[1]


def p_term_reg(p):
    "term : REGISTER"
    p[0] = functions["read_register"](p[1])


def p_expression_eq(p):
    "expression : expression EQ expression"
    p[0] = p[1] == p[3]


def p_expression_land(p):
    "expression : expression LAND expression"
    p[0] = p[1] and p[3]


def p_expression_lor(p):
    "expression : expression LOR expression"
    p[0] = p[1] or p[3]


def p_expression_lnot(p):
    "expression : LNOT expression"
    p[0] = not p[1]


def p_expression_lt(p):
    "expression : expression LT expression"
    # p[0] = p[1] < p[3]
    p[0] = Operators.ULT(p[1], p[3])


def p_expression_le(p):
    "expression : expression LE expression"
    # p[0] = p[1] <= p[3]
    p[0] = Operators.ULE(p[1], p[3])


def p_expression_gt(p):
    "expression : expression GT expression"
    # p[0] = p[1] > p[3]
    p[0] = Operators.UGT(p[1], p[3])


def p_expression_ge(p):
    "expression : expression GE expression"
    # p[0] = p[1] >= p[3]
    p[0] = Operators.UGE(p[1], p[3])


# Error rule for syntax errors
def p_error(p):
    print("Syntax error in input:", p)


# Build the parser


parser = yacc.yacc(debug=0, write_tables=0)


def parse(expression, read_memory=None, read_register=None, get_descriptor=None, word_size=32):
    global functions, sizes

    if read_memory is not None:
        functions["read_memory"] = read_memory
    else:
        functions["read_memory"] = default_read_memory

    if read_register is not None:
        functions["read_register"] = read_register
    else:
        functions["read_register"] = default_read_register

    if get_descriptor is not None:
        functions["get_descriptor"] = get_descriptor
    else:
        functions["get_descriptor"] = default_get_descriptor

    if word_size == 32:
        sizes = copy.copy(default_sizes_32)
    elif word_size == 64:
        sizes = copy.copy(default_sizes_64)
    else:
        raise ParserException("Got unsupported word size")
    result = parser.parse(expression, tracking=True)
    del functions["read_memory"]
    del functions["read_register"]
    del functions["get_descriptor"]

    return result


if __name__ == "__main__":
    while True:
        try:
            s = input("calc > ")
        except EOFError:
            break
        if not s:
            continue
        result = parse(s)
        print(result)



import io
from typing import Dict, Type, Union

from elftools.elf.elffile import ELFFile


class Binary:
    magics: Dict[bytes, Type["Binary"]] = {}

    def __new__(cls, path):
        if cls is Binary:
            with open(path, "rb") as f:
                cl = cls.magics[f.read(4)]
            return cl(path)
        else:
            return super().__new__(cls)

    def __init__(self, path):
        self.path = path
        with open(path, "rb") as f:
            self.magic = Binary.magics[f.read(4)]

    def arch(self):
        pass

    def maps(self):
        pass

    def threads(self):
        pass


class BinaryException(Exception):
    """
    Binary file exception
    """

    pass


class CGCElf(Binary):
    @staticmethod
    def _cgc2elf(filename):
        # hack begin so we can use upstream Elftool
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
            stream.write(b"\x7fELF")
            stream.name = fd.name
            return stream

    def __init__(self, filename):
        super().__init__(filename)
        stream = self._cgc2elf(filename)
        self.elf = ELFFile(stream)
        self.arch = {"x86": "i386", "x64": "amd64"}[self.elf.get_machine_arch()]

        assert "i386" == self.arch
        assert self.elf.header.e_type in ["ET_EXEC"]

    def maps(self):
        for elf_segment in self.elf.iter_segments():
            if elf_segment.header.p_type not in ["PT_LOAD", "PT_NULL", "PT_PHDR", "PT_CGCPOV2"]:
                raise BinaryException("Not Supported Section")

            if elf_segment.header.p_type != "PT_LOAD" or elf_segment.header.p_memsz == 0:
                continue

            flags = elf_segment.header.p_flags
            # PF_X 0x1 Execute - PF_W 0x2 Write - PF_R 0x4 Read
            perms = ["   ", "  x", " w ", " wx", "r  ", "r x", "rw ", "rwx"][flags & 7]
            if "r" not in perms:
                raise BinaryException("Not readable map from cgc elf not supported")

            # CGCMAP--
            assert elf_segment.header.p_filesz != 0 or elf_segment.header.p_memsz != 0
            yield (
                (
                    elf_segment.header.p_vaddr,
                    elf_segment.header.p_memsz,
                    perms,
                    elf_segment.stream.name,
                    elf_segment.header.p_offset,
                    elf_segment.header.p_filesz,
                )
            )

    def threads(self):
        yield (("Running", {"EIP": self.elf.header.e_entry}))


class Elf(Binary):
    def __init__(self, filename):
        super().__init__(filename)
        self.elf = ELFFile(open(filename, "rb"))
        self.arch = {"x86": "i386", "x64": "amd64"}[self.elf.get_machine_arch()]
        assert self.elf.header.e_type in ["ET_DYN", "ET_EXEC", "ET_CORE"]

        # Get interpreter elf
        self.interpreter = None
        for elf_segment in self.elf.iter_segments():
            if elf_segment.header.p_type != "PT_INTERP":
                continue
            self.interpreter = Elf(elf_segment.data()[:-1])
            break
        if self.interpreter is not None:
            assert self.interpreter.arch == self.arch
            assert self.interpreter.elf.header.e_type in ["ET_DYN", "ET_EXEC"]

    def __del__(self):
        if self.elf is not None:
            self.elf.stream.close()

    def maps(self):
        for elf_segment in self.elf.iter_segments():
            if elf_segment.header.p_type != "PT_LOAD" or elf_segment.header.p_memsz == 0:
                continue

            flags = elf_segment.header.p_flags
            # PF_X 0x1 Execute - PF_W 0x2 Write - PF_R 0x4 Read
            perms = ["   ", "  x", " w ", " wx", "r  ", "r x", "rw ", "rwx"][flags & 7]
            if "r" not in perms:
                raise BinaryException("Not readable map from cgc elf not supported")

            # CGCMAP--
            assert elf_segment.header.p_filesz != 0 or elf_segment.header.p_memsz != 0
            yield (
                (
                    elf_segment.header.p_vaddr,
                    elf_segment.header.p_memsz,
                    perms,
                    elf_segment.stream.name,
                    elf_segment.header.p_offset,
                    elf_segment.header.p_filesz,
                )
            )

    def getInterpreter(self):
        return self.interpreter

    def threads(self):
        yield (("Running", {"EIP": self.elf.header.e_entry}))


Binary.magics = {b"\x7fCGC": CGCElf, b"\x7fELF": Elf}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="state.proto",
    package="mserialize",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0bstate.proto\x12\nmserialize"\x1d\n\nLogMessage\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\t"\xb6\x01\n\x05State\x12\n\n\x02id\x18\x02 \x01(\x05\x12)\n\x04type\x18\x03 \x01(\x0e\x32\x1b.mserialize.State.StateType\x12\x0e\n\x06reason\x18\x04 \x01(\t\x12\x15\n\rnum_executing\x18\x05 \x01(\x05\x12\x11\n\twait_time\x18\x06 \x01(\x05"<\n\tStateType\x12\t\n\x05READY\x10\x00\x12\x08\n\x04\x42USY\x10\x01\x12\n\n\x06KILLED\x10\x02\x12\x0e\n\nTERMINATED\x10\x03".\n\tStateList\x12!\n\x06states\x18\x07 \x03(\x0b\x32\x11.mserialize.State"7\n\x0bMessageList\x12(\n\x08messages\x18\x08 \x03(\x0b\x32\x16.mserialize.LogMessageb\x06proto3',
)


_STATE_STATETYPE = _descriptor.EnumDescriptor(
    name="StateType",
    full_name="mserialize.State.StateType",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="READY",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="BUSY",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="KILLED",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TERMINATED",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=181,
    serialized_end=241,
)
_sym_db.RegisterEnumDescriptor(_STATE_STATETYPE)


_LOGMESSAGE = _descriptor.Descriptor(
    name="LogMessage",
    full_name="mserialize.LogMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="content",
            full_name="mserialize.LogMessage.content",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=27,
    serialized_end=56,
)


_STATE = _descriptor.Descriptor(
    name="State",
    full_name="mserialize.State",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="mserialize.State.id",
            index=0,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="type",
            full_name="mserialize.State.type",
            index=1,
            number=3,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reason",
            full_name="mserialize.State.reason",
            index=2,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="num_executing",
            full_name="mserialize.State.num_executing",
            index=3,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="wait_time",
            full_name="mserialize.State.wait_time",
            index=4,
            number=6,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _STATE_STATETYPE,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=59,
    serialized_end=241,
)


_STATELIST = _descriptor.Descriptor(
    name="StateList",
    full_name="mserialize.StateList",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="states",
            full_name="mserialize.StateList.states",
            index=0,
            number=7,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=243,
    serialized_end=289,
)


_MESSAGELIST = _descriptor.Descriptor(
    name="MessageList",
    full_name="mserialize.MessageList",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="messages",
            full_name="mserialize.MessageList.messages",
            index=0,
            number=8,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=291,
    serialized_end=346,
)

_STATE.fields_by_name["type"].enum_type = _STATE_STATETYPE
_STATE_STATETYPE.containing_type = _STATE
_STATELIST.fields_by_name["states"].message_type = _STATE
_MESSAGELIST.fields_by_name["messages"].message_type = _LOGMESSAGE
DESCRIPTOR.message_types_by_name["LogMessage"] = _LOGMESSAGE
DESCRIPTOR.message_types_by_name["State"] = _STATE
DESCRIPTOR.message_types_by_name["StateList"] = _STATELIST
DESCRIPTOR.message_types_by_name["MessageList"] = _MESSAGELIST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogMessage = _reflection.GeneratedProtocolMessageType(
    "LogMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _LOGMESSAGE,
        "__module__": "state_pb2"
        # @@protoc_insertion_point(class_scope:mserialize.LogMessage)
    },
)
_sym_db.RegisterMessage(LogMessage)

State = _reflection.GeneratedProtocolMessageType(
    "State",
    (_message.Message,),
    {
        "DESCRIPTOR": _STATE,
        "__module__": "state_pb2"
        # @@protoc_insertion_point(class_scope:mserialize.State)
    },
)
_sym_db.RegisterMessage(State)

StateList = _reflection.GeneratedProtocolMessageType(
    "StateList",
    (_message.Message,),
    {
        "DESCRIPTOR": _STATELIST,
        "__module__": "state_pb2"
        # @@protoc_insertion_point(class_scope:mserialize.StateList)
    },
)
_sym_db.RegisterMessage(StateList)

MessageList = _reflection.GeneratedProtocolMessageType(
    "MessageList",
    (_message.Message,),
    {
        "DESCRIPTOR": _MESSAGELIST,
        "__module__": "state_pb2"
        # @@protoc_insertion_point(class_scope:mserialize.MessageList)
    },
)
_sym_db.RegisterMessage(MessageList)


# @@protoc_insertion_point(module_scope)



from ..utils import install_helper

install_helper.ensure_native_deps()

# Exports (for `from manticore.native import ...`)
from .manticore import Manticore
from .models import variadic
from . import cpu



import copy
import logging

from typing import List, Tuple, Sequence, Optional, Any

from .smtlib import solver, Bool, issymbolic, BitVecConstant
from .smtlib.expression import Expression
from ..utils.event import Eventful
from ..utils.helpers import PickleSerializer
from ..utils import config
from .plugin import StateDescriptor

consts = config.get_group("core")
consts.add(
    "execs_per_intermittent_cb",
    default=1000,
    description="How often to fire the `exec_intermittent` event",
)

logger = logging.getLogger(__name__)


class StateException(Exception):
    """All state related exceptions"""


class TerminateState(StateException):
    """Terminates current state exploration"""

    def __init__(self, message, testcase=False):
        super().__init__(message)
        self.testcase = testcase


class AbandonState(TerminateState):
    """Exception returned for abandoned states when
    execution is finished
    """

    def __init__(self, message="Abandoned state"):
        super().__init__(message)


class Concretize(StateException):
    """Base class for all exceptions that trigger the concretization
    of a symbolic expression

    This will fork the state using a pre-set concretization policy
    Optional `setstate` function set the state to the actual concretized value.
    #Fixme Doc.

    """

    _ValidPolicies = [
        "MIN",
        "MAX",
        "MINMAX",
        "ALL",
        "SAMPLED",
        "ONE",
        "PESSIMISTIC",
        "OPTIMISTIC",
        "EXPLICIT",
    ]

    def __init__(
        self,
        message,
        expression,
        setstate=None,
        policy=None,
        values: Optional[List[Any]] = None,
        **kwargs,
    ):
        if policy is None:
            policy = "ALL"
        if policy not in self._ValidPolicies:
            raise StateException(
                f'Policy ({policy}) must be one of: {", ".join(self._ValidPolicies)}'
            )
        self.expression = expression
        self.setstate = setstate
        self.policy = policy
        self.values = values
        self.message = f"Concretize: {message} (Policy: {policy})"
        super().__init__()


class SerializeState(Concretize):
    """Allows the user to save a copy of the current state somewhere on the
    disk so that analysis can later be resumed from this point.
    """

    def _setstate(self, state, _value):
        with open(self.filename, "wb") as statef:
            PickleSerializer().serialize(state, statef)

    def __init__(self, filename, **kwargs):
        super().__init__(
            f"Saving state to {filename}",
            BitVecConstant(size=32, value=0),
            setstate=self._setstate,
            policy="ONE",
            **kwargs,
        )
        self.filename = filename


class ForkState(Concretize):
    """Specialized concretization class for Bool expressions.
    It tries True and False as concrete solutions. /

    Note: as setstate is None the concrete value is not written back
    to the state. So the expression could still by symbolic(but constrained)
    in forked states.
    """

    def __init__(self, message, expression: Bool, **kwargs):
        assert isinstance(expression, Bool), "Need a Bool to fork a state in two states"
        super().__init__(message, expression, policy="ALL", **kwargs)


class EventSolver(Eventful):
    """
    Wrapper around the solver that raises `will_solve` and `did_solve` around every call. Each call expands to:
    ```
    def method_name(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "method_name")
        solved = SelectedSolver.instance().method_name(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "method_name", solved)
        return solved
    ```
    """

    _published_events = {"solve"}

    @property
    def _solver(self):
        from .smtlib import SelectedSolver

        return SelectedSolver.instance()  # solver

    def can_be_true(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "can_be_true")
        solved = self._solver.can_be_true(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "can_be_true", solved)
        return solved

    def get_all_values(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "get_all_values")
        solved = self._solver.get_all_values(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "get_all_values", solved)
        return solved

    def get_value(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "get_value")
        solved = self._solver.get_value(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "get_value", solved)
        return solved

    def max(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "max")
        solved = self._solver.max(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "max", solved)
        return solved

    def min(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "min")
        solved = self._solver.min(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "min", solved)
        return solved

    def minmax(self, constraints, expression, *args, **kwargs):
        self._publish("will_solve", constraints, expression, "minmax")
        solved = self._solver.minmax(constraints, expression, *args, **kwargs)
        self._publish("did_solve", constraints, expression, "minmax", solved)
        return solved

    def __getattr__(self, item: str):
        """
        Pass through any undefined attribute lookups to the underlying solver
        :param item: The name of the field to get
        :return: The item, if present on self._solver
        """
        return getattr(self._solver, item)


class StateBase(Eventful):
    """
    Representation of a unique program state/path.

    :param ConstraintSet constraints: Initial constraints
    :param Platform platform: Initial operating system state
    :ivar dict context: Local context for arbitrary data storage
    """

    _published_events = {"execution_intermittent"}

    def __init__(self, constraints, platform, **kwargs):
        super().__init__(**kwargs)
        self.manticore = None
        self._platform = platform
        self._constraints = constraints
        self._platform.constraints = constraints
        self._input_symbols = list()
        self._child = None
        self._context = dict()
        self._terminated_by = None
        self._solver = EventSolver()
        self._total_exec = 0
        self._own_exec = 0
        # 33
        # Events are lost in serialization and fork !!
        self.forward_events_from(self._solver)
        self.forward_events_from(platform)

    def __getstate__(self):
        state = super().__getstate__()
        state["platform"] = self._platform
        state["constraints"] = self._constraints
        state["input_symbols"] = self._input_symbols
        state["child"] = self._child
        state["context"] = self._context
        state["terminated_by"] = self._terminated_by
        state["exec_counter"] = self._total_exec
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._platform = state["platform"]
        self._constraints = state["constraints"]
        self._input_symbols = state["input_symbols"]
        self._child = state["child"]
        self._context = state["context"]
        self._terminated_by = state["terminated_by"]
        self._total_exec = state["exec_counter"]
        self._own_exec = 0
        self._solver = EventSolver()
        # 33
        # Events are lost in serialization and fork !!
        self.forward_events_from(self._solver)
        self.forward_events_from(self._platform)

    @property
    def id(self):
        return getattr(self, "_id", None)

    def __repr__(self):
        return f"<State object with id {self.id}>"

    # Fixme(felipe) change for with "state.cow_copy() as st_temp":.
    # This need to change. this is the center of ALL the problems. re. CoW
    def __enter__(self):
        assert self._child is None
        self._platform.constraints = None
        new_state = self.__class__(self._constraints.__enter__(), self._platform)
        self.platform.constraints = new_state.constraints
        new_state._input_symbols = list(self._input_symbols)
        new_state._context = copy.copy(self._context)
        new_state._id = None
        new_state.manticore = self.manticore
        new_state._total_exec = self._total_exec
        self.copy_eventful_state(new_state)

        self._child = new_state
        assert new_state.platform.constraints is new_state.constraints

        return new_state

    def __exit__(self, ty, value, traceback):
        self._constraints.__exit__(ty, value, traceback)
        self._child = None
        self.platform.constraints = self.constraints

    @property
    def input_symbols(self):
        return self._input_symbols

    @property
    def context(self):
        return self._context

    @property
    def platform(self):
        return self._platform

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, constraints):
        self._constraints = constraints
        self.platform.constraints = constraints

    def _update_state_descriptor(self, descriptor: StateDescriptor, *args, **kwargs):
        """
        Called on execution_intermittent to update the descriptor for this state. This is intended for information
        like the PC or instruction count, where updating after each instruction would be a waste of cycles.
        This one updates the execution counts

        :param descriptor: StateDescriptor for this state
        """
        descriptor.total_execs = self._total_exec
        descriptor.own_execs = self._own_exec

    def execute(self):
        self._total_exec += 1
        self._own_exec += 1

        if self._total_exec % consts.execs_per_intermittent_cb == 0:
            self._publish("on_execution_intermittent", self._update_state_descriptor)

    def constrain(self, constraint):
        """Constrain state.

        :param manticore.core.smtlib.Bool constraint: Constraint to add
        """
        constraint = self.migrate_expression(constraint)
        self._constraints.add(constraint)

    def abandon(self):
        """Abandon the currently-active state.

        Note: This must be called from the Executor loop, or a :func:`~manticore.Manticore.hook`.
        """
        raise AbandonState

    def new_symbolic_buffer(self, nbytes, **options):
        """Create and return a symbolic buffer of length `nbytes`. The buffer is
        not written into State's memory; write it to the state's memory to
        introduce it into the program state.

        :param int nbytes: Length of the new buffer
        :param str label: (keyword arg only) The label to assign to the buffer
        :param bool cstring: (keyword arg only) Whether or not to enforce that the buffer is a cstring
                 (i.e. no NULL bytes, except for the last byte). (bool)
        :param taint: Taint identifier of the new buffer
        :type taint: tuple or frozenset

        :return: :class:`~manticore.core.smtlib.expression.Expression` representing the buffer.
        """
        label = options.get("label")
        avoid_collisions = False
        if label is None:
            label = "buffer"
            avoid_collisions = True
        taint = options.get("taint", frozenset())
        expr = self._constraints.new_array(
            name=label,
            index_max=nbytes,
            value_bits=8,
            taint=taint,
            avoid_collisions=avoid_collisions,
        )
        self._input_symbols.append(expr)

        if options.get("cstring", False):
            for i in range(nbytes - 1):
                self._constraints.add(expr[i] != 0)

        return expr

    def new_symbolic_value(self, nbits, label=None, taint=frozenset()):
        """Create and return a symbolic value that is `nbits` bits wide. Assign
        the value to a register or write it into the address space to introduce
        it into the program state.

        :param int nbits: The bitwidth of the value returned
        :param str label: The label to assign to the value
        :param taint: Taint identifier of this value
        :type taint: tuple or frozenset
        :return: :class:`~manticore.core.smtlib.expression.Expression` representing the value
        """
        assert nbits in (1, 4, 8, 16, 32, 64, 128, 256)
        avoid_collisions = False
        if label is None:
            label = "val"
            avoid_collisions = True

        expr = self._constraints.new_bitvec(
            nbits, name=label, taint=taint, avoid_collisions=avoid_collisions
        )
        self._input_symbols.append(expr)
        return expr

    def concretize(self, symbolic, policy, maxcount=7, explicit_values: Optional[List[Any]] = None):
        """This finds a set of solutions for symbolic using policy.

        This limits the number of solutions returned to `maxcount` to avoid
        a blowup in the state space. **This means that if there are more
        than `maxcount` feasible solutions, some states will be silently
        ignored.**
        """
        assert self.constraints == self.platform.constraints
        symbolic = self.migrate_expression(symbolic)

        vals: List[Any] = []
        if policy == "MINMAX":
            vals = list(self._solver.minmax(self._constraints, symbolic))
        elif policy == "MAX":
            vals = [self._solver.max(self._constraints, symbolic)]
        elif policy == "MIN":
            vals = [self._solver.min(self._constraints, symbolic)]
        elif policy == "SAMPLED":
            m, M = self._solver.minmax(self._constraints, symbolic)
            vals += [m, M]
            if M - m > 3:
                if self._solver.can_be_true(self._constraints, symbolic == (m + M) // 2):
                    vals.append((m + M) // 2)
            if M - m > 100:
                for i in (0, 1, 2, 5, 32, 64, 128, 320):
                    if self._solver.can_be_true(self._constraints, symbolic == m + i):
                        vals.append(m + i)
                    if maxcount <= len(vals):
                        break
            if M - m > 1000 and maxcount > len(vals):
                vals += self._solver.get_all_values(
                    self._constraints, symbolic, maxcnt=maxcount - len(vals), silent=True
                )
        elif policy == "ONE":
            vals = [self._solver.get_value(self._constraints, symbolic)]
        elif policy == "OPTIMISTIC":
            logger.info("Optimistic case when forking")
            if self._solver.can_be_true(self._constraints, symbolic):
                vals = [True]
            else:
                # We assume the path constraint was feasible to begin with
                vals = [False]
        elif policy == "PESSIMISTIC":
            logger.info("Pessimistic case when forking")
            if self._solver.can_be_true(self._constraints, symbolic == False):
                vals = [False]
            else:
                # We assume the path constraint was feasible to begin with
                vals = [True]
        elif policy == "EXPLICIT":
            if explicit_values:
                for val in explicit_values:
                    if self._solver.can_be_true(self._constraints, val == symbolic):
                        vals.append(val)
                    if len(vals) >= maxcount:
                        break
        else:
            assert policy == "ALL"
            vals = self._solver.get_all_values(
                self._constraints, symbolic, maxcnt=maxcount, silent=True
            )

        return tuple(set(vals))

    def migrate_expression(self, expression):
        if not issymbolic(expression):
            return expression
        migration_map = self.context.get("migration_map")
        if migration_map is None:
            migration_map = {}
        migrated_expression = self.constraints.migrate(expression, name_migration_map=migration_map)
        self.context["migration_map"] = migration_map
        return migrated_expression

    def is_feasible(self):
        return self.can_be_true(True)

    def can_be_true(self, expr):
        expr = self.migrate_expression(expr)
        return self._solver.can_be_true(self._constraints, expr)

    def can_be_false(self, expr):
        expr = self.migrate_expression(expr)
        return self._solver.can_be_true(self._constraints, expr == False)

    def must_be_true(self, expr):
        expr = self.migrate_expression(expr)
        return self._solver.can_be_true(self._constraints, expr) and not self._solver.can_be_true(
            self._constraints, expr == False
        )

    def solve_one(self, expr, constrain=False):
        """
        A version of solver_one_n for a single expression. See solve_one_n.
        """
        return self.solve_one_n(expr, constrain=constrain)[0]

    def solve_one_n(self, *exprs: Expression, constrain: bool = False) -> List[int]:
        """
        Concretize a list of symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        a list of solutions.

        :param exprs: An iterable of manticore.core.smtlib.Expression
        :param bool constrain: If True, constrain expr to solved solution value
        :return: List of concrete value or a tuple of concrete values
        """
        return self.solve_one_n_batched(exprs, constrain)

    def solve_one_n_batched(
        self, exprs: Sequence[Expression], constrain: bool = False
    ) -> List[int]:
        """
        Concretize a list of symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        a list of solutions.
        :param exprs: An iterable of manticore.core.smtlib.Expression
        :param bool constrain: If True, constrain expr to solved solution value
        :return: List of concrete value or a tuple of concrete values
        """
        # Return ret instead of value, to allow the bytearray/bytes conversion
        ret = []
        exprs = [self.migrate_expression(x) for x in exprs]
        values = self._solver.get_value_in_batch(self._constraints, exprs)
        assert len(values) == len(exprs)
        for idx, expr in enumerate(exprs):
            value = values[idx]
            if constrain:
                self.constrain(expr == values[idx])
            # Include forgiveness here
            if isinstance(value, bytearray):
                value = bytes(value)
            ret.append(value)
        return ret

    def solve_n(self, expr, nsolves):
        """
        Concretize a symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        `nsolves` solutions.

        :param manticore.core.smtlib.Expression expr: Symbolic value to concretize
        :return: Concrete value
        :rtype: list[int]
        """
        expr = self.migrate_expression(expr)
        return self._solver.get_all_values(self._constraints, expr, nsolves, silent=True)

    def solve_max(self, expr):
        """
        Solves a symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        its maximum solution

        :param manticore.core.smtlib.Expression expr: Symbolic value to solve
        :return: Concrete value
        :rtype: list[int]
        """
        if isinstance(expr, int):
            return expr
        expr = self.migrate_expression(expr)
        return self._solver.max(self._constraints, expr)

    def solve_min(self, expr):
        """
        Solves a symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        its minimum solution

        :param manticore.core.smtlib.Expression expr: Symbolic value to solve
        :return: Concrete value
        :rtype: list[int]
        """
        if isinstance(expr, int):
            return expr
        expr = self.migrate_expression(expr)
        return self._solver.min(self._constraints, expr)

    def solve_minmax(self, expr):
        """
        Solves a symbolic :class:`~manticore.core.smtlib.expression.Expression` into
        its minimum and maximun solution. Only defined for bitvects.

        :param manticore.core.smtlib.Expression expr: Symbolic value to solve
        :return: Concrete value
        :rtype: list[int]
        """
        if isinstance(expr, int):
            return expr
        expr = self.migrate_expression(expr)
        return self._solver.minmax(self._constraints, expr)

    ################################################################################################
    # The following should be moved to specific class StatePosix?

    def solve_buffer(self, addr, nbytes, constrain=False):
        """
        Reads `nbytes` of symbolic data from a buffer in memory at `addr` and attempts to
        concretize it

        :param int address: Address of buffer to concretize
        :param int nbytes: Size of buffer to concretize
        :param bool constrain: If True, constrain the buffer to the concretized value
        :return: Concrete contents of buffer
        :rtype: list[int]
        """
        buffer = self.cpu.read_bytes(addr, nbytes)
        result = []
        with self._constraints as temp_cs:
            cs_to_use = self.constraints if constrain else temp_cs
            for c in buffer:
                result.append(self._solver.get_value(cs_to_use, c))
                cs_to_use.add(c == result[-1])
        return result

    def symbolicate_buffer(
        self, data, label="INPUT", wildcard="+", string=False, taint=frozenset()
    ):
        """Mark parts of a buffer as symbolic (demarked by the wildcard byte)

        :param str data: The string to symbolicate. If no wildcard bytes are provided,
                this is the identity function on the first argument.
        :param str label: The label to assign to the value
        :param str wildcard: The byte that is considered a wildcard
        :param bool string: Ensure bytes returned can not be NULL
        :param taint: Taint identifier of the symbolicated data
        :type taint: tuple or frozenset

        :return: If data does not contain any wildcard bytes, data itself. Otherwise,
            a list of values derived from data. Non-wildcard bytes are kept as
            is, wildcard bytes are replaced by Expression objects.
        """
        if wildcard in data:
            size = len(data)
            symb = self._constraints.new_array(
                name=label, index_max=size, taint=taint, avoid_collisions=True
            )
            self._input_symbols.append(symb)

            tmp = []
            for i in range(size):
                if data[i] == wildcard:
                    tmp.append(symb[i])
                else:
                    tmp.append(data[i])

            data = tmp

        if string:
            for b in data:
                if issymbolic(b):
                    self._constraints.add(b != 0)
                else:
                    assert b != 0
        return data
